/*
  grid2cs.cpp 
  Copyright (C) 2000 by W.C.A. Wijngaards
  Copyright (C) 2003 TURBO J (XML-parts, proctex)
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include "cssysdef.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "csgeom/vector3.h"
#include "csgeom/vector2.h"
#include "csutil/cscolor.h"
#include "csutil/util.h"
#include "csutil/indprint.h"
#include "csutil/scanstr.h"
#include "csutil/csstring.h"
#include "csutil/xmltiny.h"
#include "iutil/document.h"
#include "iutil/objreg.h"
#include "iutil/vfs.h"
#include "iutil/cmdline.h"
#include "iutil/strvec.h"
#include "cstool/initapp.h"
#include "ivaria/reporter.h"
#include "readcfg.h"
#include "grid2cs.h"
#include "dochelp.h"
///--------- Global functions ----------------------------------------------
#ifndef MIN
#define MIN(x,y) (((x)<(y))?(x):(y))
#endif
#ifndef MAX
#define MAX(x,y) (((x)<(y))?(y):(x))
#endif
#ifndef ABS
#define ABS(x) ( ((x)<0)?-(x):(x) )
#endif

CS_IMPLEMENT_APPLICATION

csRef<iDocumentSystem> xml;
csRef<iDocument> doc;
csRef<iDocumentNode> root;
csRef<iVFS> vfs;

// Windows: need _snprintf
#ifdef COMP_VC 
#define snprintf _snprintf
#endif

#undef USE_OLD

#ifndef USE_OLD
#define fprintf dummy
#endif
int dummy (...) {
  return 0;
}

void synopsis(char *programname)
{
  printf("Usage: %s <input file> <output file>\n", programname);
  exit(1);
}

void CheckString(char *inputbuf, char *constantstring)
{
  if(strcmp(inputbuf, constantstring) != 0)
  {
    printf("Error: expected '%s' but found '%s'.\n", constantstring, inputbuf);
    exit(1);
  }
}


double ReadFloat(FILE *in)
{
  double b= 0.0;
  if(!ReadFloat(in, &b))
  {
    printf("Error: expected a number but found '%s'.\n", GetToken(in));
    exit(1);
  }
  return b;
}


int ReadLookID(FILE *in, int acceptspace)
{
  /// read a look ID character or integer. If acceptspace, ' ' can be read.
  /// if acceptspace then a newline will return spaces here, pushing
  /// back the newline.
  int c;
  c = fgetc(in);
  if(acceptspace && ( (c == '\n') || (c == '\r') ) )
  {
    ungetc(c, in);
    return ' ';
  }
  if(acceptspace && c == ' ')
    return c;
  /// skip whitespace.
  while(isspace(c))
    c = fgetc(in);
  /// done?
  if(c == EOF)
  {
    printf("Error: expected lookID but encountered end of file.\n");
    exit(1);
  }
  if(!isdigit(c))
    return c;
  //// c is a digit, read a number
  int num = 0;
  while(isdigit(c))
  {
    num *= 10;
    num += c - '0';
    c = fgetc(in);
  }
  if(c == EOF)
  {
    printf("Error: expected integer lookID but encountered end of file.\n");
    exit(1);
  }
  if(acceptspace && c != ' ')
  {
    printf("Error: expected space after integer but found '%c'.\n", c);
    exit(1);
  }
  return c;
}


static grVec start_loc;
static char *start_sector = 0;
void SetStart(char *secname, grVec *v)
{
  if(start_sector) free(start_sector);
  start_sector = strdup(secname);
  start_loc.x = v->x;
  start_loc.y = v->y;
  start_loc.z = v->z;
}
void GetStart(char *&secname, grVec *v)
{
  if(start_sector == 0) 
  {
    printf("No start location given.\n");
    exit(1);
  }
  secname = start_sector;
  v->x = start_loc.x;
  v->y = start_loc.y;
  v->z = start_loc.z;
}


//------------------ grTexIno ------------------------------------

grTexInfo :: grTexInfo(FILE *in)
{
  first = 0;
  amt = 0;

  grTex *tex, *last;
  ReadConstantString(in, "Textures");
  char *buf;
  buf = GetToken(in);
  last = 0;
  while( strcmp(buf, "EndTextures") != 0)
  {
    //  CheckString(buf, "tex");
    /// new item in list
    tex = new grTex;
    if (strcmp (buf,"tex")==0){
      tex->type=0;
    } else {
      if (strcmp(buf,"proctex")==0){
	tex->type=1;
      } else {
	printf("Expected tex or proctex, found %s\n",buf);
	exit(1);
      }
    }  
    tex->name = 0;
    tex->file = 0;
    tex->next = 0;
    if(last == 0) first = tex;
    else last->next = tex;
    last = tex;

    // put info in tex
    tex->name = strdup(GetToken(in));
    tex->file = strdup(GetToken(in));

    amt ++;

    buf = GetToken(in);
  }
  /// textures read correctly
  printf("Read %d texture names.\n", amt);
}


grTexInfo :: ~grTexInfo()
{
  grTex *p, *np;
  p = first;
  while(p)
  {
    np = p->next;
    delete p;
    p = np;
  }
}


void grTexInfo :: Print(FILE *out)
{
  fprintf(out, "  TEXTURES (\n");
  grTex *tex = first;
  while(tex)
  {
    fprintf(out, "    TEXTURE '%s' (FILE (%s))\n", tex->name, tex->file);
    tex = tex->next;
  }
  fprintf(out, "  )\n\n");
  fprintf(out, "  MATERIALS (\n");
  tex = first;
  while(tex)
  {
    fprintf(out, "    MATERIAL '%s' (TEXTURE(%s))\n", tex->name, tex->name);
    tex = tex->next;
  }
  fprintf(out, "  )\n\n");

}


//-------------- grLooksInfo --------------------------------------------

grLooksInfo :: grLooksInfo (FILE *in)
{
  amt = 0;
  first = 0;

  grLook *look, *last;
  ReadConstantString(in, "Looks");
  char *buf;
  buf = GetToken(in);
  last = 0;
  while( strcmp(buf, "EndLooks") != 0)
  {
    CheckString(buf, "look");
    
    /// new item in list
    look = new grLook(in);
    look->SetNext(0);
    if(last == 0) first = look;
    else last->SetNext(look);
    last = look;

    amt ++;

    buf = GetToken(in);
  }
  /// looks read correctly
  printf("Read %d looks.\n", amt);
}


grLooksInfo :: ~grLooksInfo()
{
  grLook *p, *np;
  p = first;
  while(p)
  {
    np = p->GetNext();
    delete p;
    p = np;
  }
}


grLook *grLooksInfo :: GetLook(int id)
{
  /// stupid search algo.
  /// simply making binary search in an array would be fast as hell,
  /// but unnecessary for now.
  grLook *l = first;
  while(l)
  {
    if(l->GetLookID() == id)
      return l;
    l = l->GetNext();
  }
  if(id<128)
    printf("Error: reference made to undefined look %c (%d).\n", id, id);
  else printf("Error: reference made to undefined look %d.\n", id);
  exit(1);
  return 0;
}


//---------- grLook ------------------------------------------------
grLook :: grLook (FILE *in)
{
  /// init
  lookid = 0;
  floorheight = 0.0; 
  ceilingheight = 10.0;
  walltex = "white.gif";
  floortex = "white.gif";
  ceiltex = "white.gif";
  /// light properties
  has_light = 0; // no light
  radius = 5.0;
  r = 1.0;
  g = 1.0;
  b = 1.0;
  lightheight = 1.0;
  /// start properties
  is_start = 0;

  /// read
  /// first read mandatory part, then try for optional addons.
  lookid = ReadLookID(in, false);
  ReadConstantString(in, "floor"); floorheight = ReadFloat(in);
  ReadConstantString(in, "ceiling"); ceilingheight = ReadFloat(in);
  ReadConstantString(in, "walltex"); walltex = strdup(GetToken(in));
  ReadConstantString(in, "floortex"); floortex = strdup(GetToken(in));
  ReadConstantString(in, "ceiltex"); ceiltex = strdup(GetToken(in));

  /// optional addons
  char *buf;
  buf = GetToken(in);
  while( strcmp(buf, "end") != 0)
  {
    if(strcmp(buf, "light")==0)
    {
      /// read light info
      /// multiple light infos will overwrite eachother
      has_light = true;
      ReadConstantString(in, "radius"); radius = ReadFloat(in);
      ReadConstantString(in, "color"); r = ReadFloat(in);
      g = ReadFloat(in); b = ReadFloat(in);
      ReadConstantString(in, "lightheight"); lightheight = ReadFloat(in);
      ReadConstantString(in, "endlight");
    }
    else if(strcmp(buf, "dlight")==0)
    {
      /// read light info
      /// multiple light infos will overwrite eachother
      has_light = 2;
      ReadConstantString(in, "radius"); radius = ReadFloat(in);
      ReadConstantString(in, "color"); r = ReadFloat(in);
      g = ReadFloat(in); b = ReadFloat(in);
      ReadConstantString(in, "lightheight"); lightheight = ReadFloat(in);
      ReadConstantString(in, "endlight");
    }
    else if(strcmp(buf, "start")==0)
    {
      is_start = 1;
    }
    else 
    {
      // error
      printf("Error: expected option or end of look but found '%s'.\n", buf);
      exit(1);
    }
    buf = GetToken(in);
  }
  // read correctly.
}


grLook :: ~grLook()
{
  // nothing to do
}


void grLook :: Polygonize(grSector *sec, int x, int y)
{
  grPolSet *pols = sec->GetPolSet();
  grTexPlaneInfo *texplanes = sec->GetSectorInfo()->GetTexPlaneInfo();
  grMap *map = sec->GetSectorInfo()->GetMap();
  /// create floor poly
  int v1,v2,v3,v4;
  float fx = x;
  float fy = y;
  grVec orig(-0.5, floorheight, -0.5);
  grVec uvec(0.5, floorheight, -0.5);
  grVec vvec(-0.5, floorheight, 0.5);
  grTexPlane *plane = texplanes->AddPlane(floortex, &orig, &uvec, &vvec);
  v1 = pols->AddVertice( map->Get3dPos(fx-0.5, fy-0.5, floorheight));
  v2 = pols->AddVertice( map->Get3dPos(fx+0.5, fy-0.5, floorheight));
  v3 = pols->AddVertice( map->Get3dPos(fx+0.5, fy+0.5, floorheight));
  v4 = pols->AddVertice( map->Get3dPos(fx-0.5, fy+0.5, floorheight));
  pols->AddPolygon(v1, v2, v3, v4, floortex, plane);

  /// ceiling polygon
  orig.Set(-0.5, ceilingheight, 0.5);
  uvec.Set(0.5, ceilingheight, 0.5);
  vvec.Set(-0.5, ceilingheight, -0.5);
  plane = texplanes->AddPlane(ceiltex, &orig, &uvec, &vvec);
  v1 = pols->AddVertice( map->Get3dPos(fx-0.5, fy+0.5, ceilingheight));
  v2 = pols->AddVertice( map->Get3dPos(fx+0.5, fy+0.5, ceilingheight));
  v3 = pols->AddVertice( map->Get3dPos(fx+0.5, fy-0.5, ceilingheight));
  v4 = pols->AddVertice( map->Get3dPos(fx-0.5, fy-0.5, ceilingheight));
  pols->AddPolygon(v1, v2, v3, v4, ceiltex, plane);

  /// walls...
  /// north wall
  addwall(sec, x, y, x, y-1, 1, 0, 0.0,-0.5);
  /// east wall
  addwall(sec, x, y, x+1, y, 0, 1, 0.5, 0.0);
  /// south wall
  addwall(sec, x, y, x, y+1,-1, 0, 0.0, 0.5);
  /// west wall
  addwall(sec, x, y, x-1, y, 0,-1,-0.5, 0.0);

}


void grLook :: addwall(grSector *sec, 
  int x, int y, int neighbor_x, int neighbor_y,
      int dir_x, int dir_y, float shift_x, float shift_y)
{
  grPolSet *pols = sec->GetPolSet();
  grTexPlaneInfo *texplanes = sec->GetSectorInfo()->GetTexPlaneInfo();
  grMap *map = sec->GetSectorInfo()->GetMap();

  /// top polygon heights
  float top_max, top_min; 
  /// bottom polygon heights
  float bot_max, bot_min; 
  /// make 2 polygons?
  float make_two = 0;
  if( map->IsReachable(neighbor_x, neighbor_y) )
  {
    make_two = 1;
    /// top (max == min could happen)
    top_max=ceilingheight;
    top_min=MIN(ceilingheight, map->GetCeilingHeight(neighbor_x, neighbor_y));
    /// bottom (max == min could happen)
    bot_max=MAX( floorheight, map->GetFloorHeight(neighbor_x, neighbor_y) );
    bot_min=floorheight;
  }
  else 
  {
    // neighbor can't be reached make only one polygon
    // the 'top' one
    make_two = 0;
    top_max = ceilingheight;
    top_min = floorheight;
  }
  /// make top polygon
  grTexPlane *plane = 0;
  float fx = float(x) + shift_x;
  float fy = float(y) + shift_y;
  float fdir_x = dir_x/2.0;
  float fdir_y = dir_y/2.0;
  int v1,v2,v3,v4;
  if(top_max - top_min > 0.0001) // sanity check
  {
    plane = texplanes->AddPlane(walltex, map->Get3dPos(fx,fy,top_max),
      dir_x, dir_y);
    v1 = pols->AddVertice( map->Get3dPos(fx-fdir_x, fy-fdir_y, top_max));
    v2 = pols->AddVertice( map->Get3dPos(fx+fdir_x, fy+fdir_y, top_max));
    v3 = pols->AddVertice( map->Get3dPos(fx+fdir_x, fy+fdir_y, top_min));
    v4 = pols->AddVertice( map->Get3dPos(fx-fdir_x, fy-fdir_y, top_min));
    pols->AddPolygon(v1, v2, v3, v4, walltex, plane);
  }
  if(!make_two) return;
  /// make bottom polygon
  if(bot_max - bot_min > 0.0001) // sanity check
  {
    plane = texplanes->AddPlane(walltex, map->Get3dPos(fx,fy,bot_max),
      dir_x, dir_y);
    v1 = pols->AddVertice( map->Get3dPos(fx-fdir_x, fy-fdir_y, bot_max));
    v2 = pols->AddVertice( map->Get3dPos(fx+fdir_x, fy+fdir_y, bot_max));
    v3 = pols->AddVertice( map->Get3dPos(fx+fdir_x, fy+fdir_y, bot_min));
    v4 = pols->AddVertice( map->Get3dPos(fx-fdir_x, fy-fdir_y, bot_min));
    pols->AddPolygon(v1, v2, v3, v4, walltex, plane);
  }
}


void grLook :: PrintInSector(int x, int y, FILE *out, grSector *sec)
{
  /// have any specials? if not, skip
  if(!has_light && !is_start)
    return;

  /// print specials inside the SECTOR statement
  grMap *map = sec->GetSectorInfo()->GetMap();

  if(has_light)
  {
    static int light_num = 0;
    grVec *center = map->Get3dPos(x,y,lightheight);
    fprintf(out, "    LIGHT 'light%d' (CENTER(%g,%g,%g) RADIUS(%g) "
      "COLOR(%g,%g,%g))\n", light_num++,
      center->x, center->y, center->z,
      radius, r, g, b);
  }

  if(is_start)
  {
    SetStart(sec->GetName(), map->Get3dPos(x,y,1.8));
  }

}

void grLook :: WriteLook(int x, int y, csRef<iDocumentNode> node, grSector* sec)
{
 /// have any specials? if not, skip
  if(!has_light && !is_start)
    return;
  grMap *map = sec->GetSectorInfo()->GetMap();
  char c[100];
  float rad=radius;
  if(has_light)
  { static int light_num = 0;
    grVec *center = map->Get3dPos(x,y,lightheight);
    DocNode light=CreateNode(node,"light");
    snprintf(c,99,"light%d",light_num++);
    light->SetAttribute("name",c);
    DocNode cen=CreateNode(light,"center");
    cen->SetAttributeAsFloat("x",center->x);
    cen->SetAttributeAsFloat("y",center->y);
    cen->SetAttributeAsFloat("z",center->z);
    CreateNode(light,"radius",rad);
    DocNode col=CreateNode(light,"color");
    col->SetAttributeAsFloat("red",r);
    col->SetAttributeAsFloat("green",g);
    col->SetAttributeAsFloat("blue",b);
    if (has_light==2) {
      CreateNode(light,"dynamic");
    }
  }
  if(is_start)
  {
    SetStart(sec->GetName(), map->Get3dPos(x,y,1.75));
  }
}
//--------------- grMap -----------------------------------------------

grMap :: grMap(FILE *in)
{
  double tmp;
  map = 0;
  lookinfo = 0;
  ReadConstantString(in, "Map");
  ReadConstantString(in, "mapwidth"); width = (int)ReadFloat(in); 
  ReadConstantString(in, "mapheight"); height = (int)ReadFloat(in); 
  ReadConstantString(in, "mapcontent");

  size = width*height;
  map = new int[size];

  /// read map line by line
  char buf[100];
  fgets(buf, 90, in); // skip end of line after 'mapcontent'
  for(int y = 0; y < height ; y++)
  {
    for(int x = 0; x < width; x++)
      map[y*width + x] = ReadLookID(in, 1);
    fgets(buf, 90, in);
  }

  /// done
  ReadConstantString(in, "EndMap");
  printf("Read %dx%d map.\n", width, height);
  PrintMap();
}


grMap :: ~grMap()
{
  delete[] map;
}


void grMap :: PrintMap()
{
  for(int y=0; y<height; y++)
  {
    for(int x=0; x<width; x++)
    {
      if(map[y*width + x] > 126)
        printf("?");
      else printf("%c", map[y*width + x]);   
    }
    printf("\n");
  }
}


int grMap :: IsReachable(int x, int y)
{
  if(x < 0 || x >= width)
    return 0;
  if(y < 0 || y >= height)
    return 0;
  if(map[y*width + x] != 32) /// 32 == ' '
    return 1;
  else return 0;
}

double grMap :: GetFloorHeight(int x, int y)
{
  if(x < 0 || x >= width)
  {
    printf("Error x out of range in getFloorHeight.\n");
    exit(1);
  }
  if(y < 0 || y >= height)
  {
    printf("Error y out of range in getFloorHeight.\n");
    exit(1);
  }
  int id = GetLookID(x, y);
  grLook *l = lookinfo->GetLook(id);
  return l->GetFloorHeight();
}

double grMap :: GetCeilingHeight(int x, int y)
{
  if(x < 0 || x >= width)
  {
    printf("Error x out of range in getCeilingHeight.\n");
    exit(1);
  }
  if(y < 0 || y >= height)
  {
    printf("Error y out of range in getCeilingHeight.\n");
    exit(1);
  }
  int id = GetLookID(x, y);
  grLook *l = lookinfo->GetLook(id);
  return l->GetCeilingHeight();
}


grVec * grMap :: Get3dPos(float x, float y, float height)
{
  static grVec v;
  v.x = x * 1.0;
  v.z = -y * 1.0;
  v.y = height;
  return &v;
}


grLook * grMap :: GetLook(int x, int y)
{
  return lookinfo->GetLook(GetLookID(x, y));
}


//---------------- grSector --------------------------------------
grSector :: grSector(grSectorInfo *parent)
{
  static int next_num = 0;
  info = parent;
  sector_num = next_num++;
  pols = new grPolSet();
  x = 0;
  y = 0;
  w = 0;
  h = 0;
  next = 0;
}


grSector :: ~grSector()
{
  delete pols;
}


void grSector :: Polygonize()
{
  /// first create the portals
  /// try to make unfragmented portals
  int dir_x, dir_y; // direction to go in when making a line
  int start_x, start_y; // start position
  int todo_dist; // distance to cover
  float shift_x, shift_y; // move polygons a little.

  //printf("--- Polygonizing sector %d\n", sector_num);
  //printf("--- Polygonizing sector at %d,%d w=%d h=%d\n", x,y,w,h);
  /// make north portals
  /// work left to right
  start_x = x;
  start_y = y - 1; // look one above the current room.
  dir_x = 1;
  dir_y = 0;
  todo_dist = w;
  shift_x = 0.0;
  shift_y = -0.5 + 1.0;
  create_portals(start_x, start_y, dir_x, dir_y, todo_dist, shift_x, shift_y);

  /// create east portals
  start_x = x+w;
  start_y = y;
  dir_x = 0;
  dir_y = +1;
  todo_dist = h;
  shift_x = +0.5 - 1.0;
  shift_y = 0.0;
  create_portals(start_x, start_y, dir_x, dir_y, todo_dist, shift_x, shift_y);

  /// create south portals
  start_x = x+w -1;
  start_y = y+h;
  dir_x = -1;
  dir_y = 0;
  todo_dist = w;
  shift_x = 0.0;
  shift_y = +0.5 - 1.0;
  create_portals(start_x, start_y, dir_x, dir_y, todo_dist, shift_x, shift_y);

  /// create west portals
  start_x = x-1;
  start_y = y+h-1;
  dir_x = 0;
  dir_y = -1;
  todo_dist = h;
  shift_x = -0.5 + 1.0;
  shift_y = 0.0;
  create_portals(start_x, start_y, dir_x, dir_y, todo_dist, shift_x, shift_y);

  /// create all the nonportal polygons
  /// fragmenting the polygons (per grid point) will give a nicer
  /// result.
  int run_x, run_y;
  for(run_y = 0; run_y < h; run_y ++)
    for(run_x = 0; run_x < w; run_x ++)
    {
      // polygonize gridpoint
      grLook *lk = info->GetMap()->GetLook(run_x + x, run_y + y);
      lk->Polygonize(this, run_x + x, run_y + y);
    }

}


void grSector :: create_portals( int start_x, int start_y,
  int dir_x, int dir_y, int todo_dist, float shift_x, float shift_y)
{
  //printf("create portals (sec%d)\n", sector_num);
  //printf("create portals %d,%d %d,%d %d %g,%g\n", start_x, start_y,
  //  dir_x, dir_y, todo_dist, shift_x, shift_y);
  // gather some information from the parent.
  grMap *map = info->GetMap();
  grTexPlaneInfo *texplanes = info->GetTexPlaneInfo();
  float sector_ceiling = map->GetCeilingHeight(x,y); // should be the same
  float sector_floor = map->GetFloorHeight(x,y); // for all this sector

  int found_dist; // the distance found
  while(todo_dist > 0)
  {
    /// find a stretch that has the same sector on the other side.
    if(!info->SameSector(start_x, start_y, dir_x, dir_y, todo_dist, found_dist))
    {
      /// if none found, continue. (solid space perhaps)
      /// skip the area found
      //printf("create portals skipping %d unreachable\n", found_dist);
      start_x += found_dist * dir_x;
      start_y += found_dist * dir_y;
      todo_dist -= found_dist;
      continue;
    }
    // found a stretch of found_dist with the same sector on the other
    // side. Make a portal to it.
    // heights of portal:
    //printf("create portals portalling %d samesector\n", found_dist);
    float portal_top = MIN( map->GetCeilingHeight(start_x, start_y), 
      sector_ceiling) ;
    float portal_bottom = MAX( map->GetFloorHeight(start_x, start_y), 
      sector_floor) ;
    if(portal_top <= portal_bottom)
    {
      //printf("create portals skipping %d empty portals\n", found_dist);
      // skip this empty polygon
      start_x += found_dist * dir_x;
      start_y += found_dist * dir_y;
      todo_dist -= found_dist;
      continue;
    }
    // with vertices:
    float pol_start_x = float(start_x) - float(dir_x)/2.0 + shift_x;
    float pol_start_y = float(start_y) - float(dir_y)/2.0 + shift_y;
    float end_x = float(start_x + found_dist * dir_x) - float(dir_x)/2.0 
      + shift_x;
    float end_y = float(start_y + found_dist * dir_y) - float(dir_y)/2.0 
      + shift_y;
    grVec *v1_pos = map->Get3dPos(pol_start_x, pol_start_y, portal_top);
    int v1 = pols->AddVertice( v1_pos );
    int v2 = pols->AddVertice( map->Get3dPos(end_x, end_y, portal_top) );
    int v3 = pols->AddVertice( map->Get3dPos(end_x, end_y, portal_bottom) );
    int v4 = pols->AddVertice( map->Get3dPos(pol_start_x, pol_start_y, portal_bottom) );
    /// more poly info
    char *texture = map->GetLook(x,y)->GetWallTexture();
    char *dest_sector = info->GetSector(start_x, start_y)->GetName();
    grTexPlane *plane = texplanes->AddPlane(texture, v1_pos, dir_x, dir_y);
    pols->AddPolygon(v1, v2, v3, v4, texture, plane, dest_sector);
    //printf("Added portal from sec%d to %s\n", sector_num, dest_sector);

    start_x += found_dist * dir_x;
    start_y += found_dist * dir_y;
    todo_dist -= found_dist;
  }
}


char * grSector :: GetName()
{
  static char buf[100];
  sprintf(buf, "sec%d", sector_num);
  return buf;
}


void grSector :: Print(FILE *out)
{
  // first sector
  fprintf(out, "  SECTOR '%s' (\n", GetName());
  // for 0.19
  fprintf(out,
"   MESHOBJ 'walls' (\n"
"      ZFILL ()\n" 
"	 PLUGIN ('thing')\n"
"	 PARAMS (\n"  );
  pols->Print(out);
  
  fprintf (out,
"   ))\n");
  // then grLook specials (if any)
  for(int posy=0; posy<h; posy++)
    for(int posx=0; posx<w; posx++)
    {
      if(info->GetMap()->IsReachable(posx+x, posy+y))
        info->GetMap()->GetLook(posx+x, posy+y)->PrintInSector(
	  posx+x, posy+y, out, this);
    }
  fprintf(out, "  )\n");
}

void grSector ::WriteSector(csRef<iDocumentNode> node)
{
  DocNode sector = CreateNode (node, "sector");
  sector->SetAttribute ("name",GetName()); 
  DocNode meshobj = CreateNode (sector, "meshobj");
  meshobj->SetAttribute("name","walls");
  CreateNode (meshobj, "zfill"); 
  CreateNode (meshobj, "plugin", "thing");
  DocNode params = CreateNode (meshobj, "params");
  pols->WritePolSET(params);
  for(int posy=0; posy<h; posy++)
    for(int posx=0; posx<w; posx++)
    {
      if(info->GetMap()->IsReachable(posx+x, posy+y))
        info->GetMap()->GetLook(posx+x, posy+y)->WriteLook(
	  posx+x, posy+y, sector, this);
    }
}
//---------------- grSectorInfo ----------------------------------

grSectorInfo :: grSectorInfo()
{
  first = 0;
  map = 0;
  flags = 0;
  sector = 0;
  texplanes = 0;
  num = 0;

  nr_poly = 0;
  nr_portal = 0;
  nr_vert = 0;
}


grSectorInfo :: ~grSectorInfo()
{
  grSector *p, *np;
  p = first;
  while(p)
  {
    np = p->next;
    delete p;
    p = np;
  }
  delete[] flags;
  delete[] sector;
}


void grSectorInfo :: Sectorize(grMap *map)
{
  grSector *last = 0;
  SetMap(map);
  int x,y,w,h;
  while(findnewarea(x,y))
  {
    w=1; h=1; // start with 1x1 rectangle.
    growarea(x,y,w,h);
    grSector *sec = new grSector(this);
    sec->x = x;
    sec->y = y;
    sec->w = w;
    sec->h = h;
    if(last == 0) first = sec;
    else last->next = sec;
    last = sec;
    SetSectorized(sec, x, y, w, h);
    printf(".");
    //printf("Found sector %d: %d,%d w=%d h=%d\n", sec->sector_num, x, y, w, h);
    //PrintSectors();
    fflush(stdout);
    num++;
  }

  printf("\n");
}


void grSectorInfo :: SetMap(grMap *m)
{
  /// if(map) memory leak happens.
  map = m;
  flags = new int [map->GetSize()];
  sector = new grSector* [map->GetSize()];
  for(int i=0; i<map->GetSize(); i++)
  {
    flags[i] = 0;
    sector[i] = 0;
  }
}


int grSectorInfo :: SameSector(int start_x, int start_y, int dir_x, int dir_y, 
      int todo_dist, int& found_dist)
{
  int i;
  int cur_x = start_x;
  int cur_y = start_y;
  if(!map->IsReachable(start_x, start_y))
  {
    // a stretch of unreachable points...
    for(i=1; i<todo_dist; i++)
    {
      cur_x = start_x + i * dir_x; 
      cur_y = start_y + i * dir_y;
      if(map->IsReachable(cur_x, cur_y))
      {
        found_dist = i;
	return 0;
      }
    }
    found_dist = todo_dist;
    return 0;
  }
  /// a sector exists there.
  grSector* which_sector = GetSector(start_x, start_y);
  for(i=1; i<todo_dist; i++)
  {
    cur_x = start_x + i * dir_x; 
    cur_y = start_y + i * dir_y;
    if(GetSector(cur_x, cur_y) != which_sector)
    {
      found_dist = i;
      return 1;
    }
  }
  found_dist = todo_dist;
  return 1;
}


int grSectorInfo :: findnewarea(int &x, int &y)
{
  for(y=0; y<map->GetHeight(); y++)
    for(x=0; x<map->GetWidth(); x++)
      if(map->IsReachable(x,y) && !GetSectorized(x,y))
        return 1;
  return 0;
}


void grSectorInfo :: growarea(int x, int y, int &w, int &h)
{
  double floor = map->GetFloorHeight(x,y);
  double ceiling = map->GetCeilingHeight(x,y);
  /// grow area to maximum size.
  while( checkarea(x,y, w+1, h, floor, ceiling) )
    w++;
  while( checkarea(x,y, w, h+1, floor, ceiling) )
    h++;
}


int grSectorInfo :: checkarea(int x, int y, int w, int h, 
  double floor, double ceiling)
{
  /// look for unsectorized,
  ///   reachable,
  ///   same floor/ceiling height
  /// in the area given.
  for(int v = y; v < y+h; v++)
    for(int u = x; u < x+w; u++)
    {
      if(!map->IsReachable(u,v)) return 0;
      if(GetSectorized(u,v)) return 0;
      if(ABS(map->GetFloorHeight(u,v) - floor) > 0.00001) return 0;
      if(ABS(map->GetCeilingHeight(u,v) - ceiling) > 0.00001) return 0;
    }

}


void grSectorInfo :: SetSectorized(grSector *sec, int x, int y, int w, int h)
{
  for(int v = y; v < y+h; v++)
    for(int u = x; u < x+w; u++)
    {
      sector[v*map->GetWidth()+u] = sec;
      SetSectorized(u,v);
    }
}


void grSectorInfo :: Polygonize(grTexPlaneInfo *tplanes)
{
  texplanes = tplanes;
  grSector *sec = first;
  while(sec)
  {
    sec->Polygonize();
    nr_poly += sec->GetPolSet()->GetNumberPolygons();
    nr_portal += sec->GetPolSet()->GetNumberPortals();
    nr_vert += sec->GetPolSet()->GetNumberVertices();
    sec = sec->next;
  }
  printf("Created %d vertices for %d polygons of which %d were portals.\n",
    nr_vert, nr_poly, nr_portal);
}


void grSectorInfo :: Print(FILE *out)
{
  grSector *sec = first;
  while(sec)
  {
    sec->Print(out);
    sec = sec->next;
  }
}

void grSectorInfo ::WriteSectors(csRef<iDocumentNode> node)
{
 grSector *sec = first;
  while(sec)
  {
    sec->WriteSector(node);
    sec = sec->next;
  }
}
void grSectorInfo :: PrintSectors()
{
  int w= map->GetWidth();
  int h= map->GetHeight();

  for(int y=0; y<h; y++)
  {
    for(int x=0; x<w; x++)
    {
      grSector *sec = GetSector(x, y);
      if(!sec) printf(" ");
      else printf("%c", '0' + sec->sector_num );
    }
    printf("\n");
  }
}


//----------------------- grPolSet -------------------------------

#define POLSETINCR 20

grPolSet :: grPolSet()
{
  nr_v = 0;
  nr_p = 0;
  number_portals = 0;

  max_v = POLSETINCR;
  verts = (grVec*) malloc( sizeof(grVec) * max_v);
  first = 0;
}


grPolSet :: ~grPolSet()
{
  grPol *p, *np;
  p = first;
  while(p)
  {
    np = p->next;

    delete[] p->verts;
    if(p->texture) free(p->texture);
    // if(p->plane) free(p->plane);
    if(p->portal_dest) free(p->portal_dest);
    delete p;
    p = np;
  }
  free(verts);
}


int grPolSet :: AddVertice(grVec *vec)
{
  return AddVertice(vec->x, vec->y, vec->z);
}


int grPolSet :: AddVertice(float x, float y, float z)
{
  int i;
  // check if already present - a slow check.
  for(i=0; i < nr_v; i++)
  {
    if(  (ABS(verts[i].x - x) < 0.00001)
      && (ABS(verts[i].y - y) < 0.00001)
      && (ABS(verts[i].z - z) < 0.00001) )
      {
        return i;
      }
  }
  // not present yet, add it to the list.

  // do we need to realloc?
  if(nr_v == max_v)
  {
    max_v += POLSETINCR;
    verts = (grVec*) realloc( verts, sizeof(grVec) * max_v);
  }

  nr_v ++;
  verts[nr_v - 1].x = x;
  verts[nr_v - 1].y = y;
  verts[nr_v - 1].z = z;
  return nr_v - 1;
}


void grPolSet :: AddPolygon(int i1, int i2, int i3, int i4, 
  char *texture, grTexPlane *texplane, char *portal_dest )
{
  // add it.
  grPol *p = new grPol;
  p->amt = 4;
  p->verts = new int[4];
  p->verts[0] = i1;
  p->verts[1] = i2;
  p->verts[2] = i3;
  p->verts[3] = i4;
  p->texture = strdup(texture);
  p->plane = texplane;
  if(portal_dest)
  {
    p->portal_dest = strdup(portal_dest);
    number_portals++;
  }
  else p->portal_dest = 0;

  p->next = first;
  first = p;
  nr_p++;
}


void grPolSet :: Print(FILE *out)
{
  // print (while inside THING or SECTOR statement)
  // vertices.
  int i;
  for(i=0; i<nr_v; i++)
  {
    fprintf(out, "    VERTEX (%g, %g, %g) \n",
      verts[i].x, verts[i].y, verts[i].z);
  }
  // polygons
  grPol *pol = first;
  int nr = 0;
  while(pol)
  {
    if(pol->amt != 4)
      printf("Cannot deal with != 4 in grPolSet::Print !\n");
    fprintf(out, "    POLYGON 'p%d' (VERTICES(%d,%d,%d,%d) MATERIAL('%s') "
      "TEXTURE(PLANE('%s'))", nr++,
      pol->verts[0], pol->verts[1], pol->verts[2], pol->verts[3], 
      pol->texture, pol->plane->plane_name);
    if(pol->portal_dest != 0) // is a portal
      fprintf(out, " PORTAL('%s')", pol->portal_dest);
    fprintf(out, ")\n");
    pol = pol->next;
  }
}

void grPolSet :: WritePolSET(csRef<iDocumentNode> node)
{ int i;
 char c [100];
 
  for(i=0; i<nr_v; i++)
  {
    DocNode v = CreateNode (node, "v");
    v->SetAttributeAsFloat ("x", verts[i].x);
    v->SetAttributeAsFloat ("y", verts[i].y);
    v->SetAttributeAsFloat ("z", verts[i].z);
  }
  grPol *pol = first;
  int nr = 0; 
  while(pol)
  { nr++;
    DocNode p= CreateNode(node,"p");
    
    for (i=0 ;i<pol->amt; i++) {
      CreateNode(p,"v",(int)pol->verts[i]);
    }
    snprintf(c,99,"p%d",nr);
    p->SetAttribute("name",c);  
    CreateNode(p,"material",pol->texture);
    pol->plane->WritePlane(p);
    if (pol->portal_dest) {// portal ?!
      CreateNode(CreateNode(p,"portal"),"sector",pol->portal_dest);
    
    }
    pol=pol->next;
  }

}
//----------------------- grTexPlaneInfo -------------------------

grTexPlaneInfo :: grTexPlaneInfo()
{
  first = 0;
  amt = 0;
}


grTexPlaneInfo :: ~grTexPlaneInfo()
{
  grTexPlane *p, *np;
  p = first;
  while(p)
  {
    np = p->next;
    delete p;
    p = np;
  }
}


grTexPlane* grTexPlaneInfo :: AddPlane(char *tex, grVec *orig, grVec *v1, grVec *v2)
{
  static int plnum = 0;

  /// first look if plane is already in the list
  grTexPlane *plane = first;
  while(plane)
  {
    if( 
        ( ABS(plane->orig->x - orig->x) < 0.00001 )
     && ( ABS(plane->orig->y - orig->y) < 0.00001 )
     && ( ABS(plane->orig->z - orig->z) < 0.00001 )
     && ( ABS(plane->v1->x - v1->x) < 0.00001 )
     && ( ABS(plane->v1->y - v1->y) < 0.00001 )
     && ( ABS(plane->v1->z - v1->z) < 0.00001 )
     && ( ABS(plane->v2->x - v2->x) < 0.00001 )
     && ( ABS(plane->v2->y - v2->y) < 0.00001 )
     && ( ABS(plane->v2->z - v2->z) < 0.00001 )
     && (strcmp(plane->texture_name, tex) == 0) )
       return plane;
    plane = plane->next;
  }

  /// add a new plane
  char buf[100];
  sprintf(buf, "texplane%d", plnum++);
  grTexPlane *pl = new grTexPlane;
  pl->texture_name = strdup(tex);
  pl->plane_name = strdup(buf);
  pl->orig = new grVec;
  pl->orig->x = orig->x;
  pl->orig->y = orig->y;
  pl->orig->z = orig->z;
  pl->v1 = new grVec;
  pl->v1->x = v1->x;
  pl->v1->y = v1->y;
  pl->v1->z = v1->z;
  pl->v2 = new grVec;
  pl->v2->x = v2->x;
  pl->v2->y = v2->y;
  pl->v2->z = v2->z;
  pl->next = first;
  first = pl;
  return pl;
}


grTexPlane * grTexPlaneInfo :: AddPlane(char *tex, grVec *pt, int dir_x, int dir_y )
{
  /// the grVec is in 3d.
  const float len_1 = 1.0;
  const float len_2 = 1.0;
  grVec orig, v1,v2;
  orig.x = pt->x;
  orig.y = 0.0;
  orig.z = pt->z;
  v1.x = pt->x; // v1 will point to the right
  v1.y = 0.0;
  v1.z = pt->z;
  v2.x = pt->x; // v2 down
  v2.y = -len_2;
  v2.z = pt->z;

  if(dir_x == 0 && dir_y == 0)
  {
    printf("Cannot handle plane, dir=(0,0)\n");
    exit(1);
  }
  else if(dir_x == 0)
  {
    orig.z = 0.0;
    v2.z = 0.0;
    if(dir_y > 0) v1.z=len_1;
    if(dir_y < 0) v1.z= -len_1;
  }
  else if(dir_y == 0)
  {
    orig.x = 0.0;
    v2.x = 0.0;
    if(dir_x > 0) v1.x=len_1;
    if(dir_x < 0) v1.x=-len_1;
  }
  else {
    printf("Cannot handle plane\n");
    exit(1);
  }

  return AddPlane(tex, &orig, &v1, &v2);
}


void grTexPlaneInfo :: Print(FILE *out)
{
  grTexPlane *tp = first;
  while(tp)
  {
    // Print PLANE
    fprintf(out, "  ADDON (PLUGIN ('plane') PARAMS ( NAME ('%s') "
    "ORIG (%g,%g,%g) FIRST (%g,%g,%g) "
                 "SECOND (%g,%g,%g) FIRST_LEN(%g) SECOND_LEN(%g) ))\n", 
      tp->plane_name, 
      tp->orig->x, tp->orig->y, tp->orig->z,
      tp->v1->x, tp->v1->y, tp->v1->z,
      tp->v2->x, tp->v2->y, tp->v2->z,
      1.0, 1.0);
    tp = tp->next;
  }
}
void grTexPlane :: WritePlane(csRef<iDocumentNode> node)
{
  DocNode texmap= CreateNode(node,"texmap");
  DocNode o=CreateNode(texmap,"orig");
  o->SetAttributeAsFloat("x",orig->x);
  o->SetAttributeAsFloat("y",orig->y);
  o->SetAttributeAsFloat("z",orig->z);
  o=CreateNode(texmap,"first");
  o->SetAttributeAsFloat("x",v1->x);
  o->SetAttributeAsFloat("y",v1->y);
  o->SetAttributeAsFloat("z",v1->z);
  CreateNode(texmap,"firstlen",1);
  o=CreateNode(texmap,"second");
  o->SetAttributeAsFloat("x",v2->x);
  o->SetAttributeAsFloat("y",v2->y);
  o->SetAttributeAsFloat("z",v2->z);
  CreateNode(texmap,"secondlen",1);

}
bool WritePlugins(csRef<iDocumentNode> node)
{
    DocNode plugins=CreateNode(node,"plugins");
    DocNode plugin;
    plugin=CreateNode(plugins,"plugin","crystalspace.mesh.loader.thing");
  plugin->SetAttribute ("name", "thing");
  plugin = CreateNode (plugins, "plugin", "crystalspace.mesh.loader.factory.thing");
  plugin->SetAttribute ("name", "thingFact");

  plugin = CreateNode (plugins, "plugin", "crystalspace.mesh.loader.bezier");
  plugin->SetAttribute ("name", "bezier");
  plugin = CreateNode (plugins, "plugin", "crystalspace.mesh.loader.factory.bezier");
  plugin->SetAttribute ("name", "bezierFact");

  return true;
}

void grTexInfo::WriteTextures(csRef<iDocumentNode> node)

{
    DocNode textures=CreateNode (node,"textures");
    grTex *tex=first;
    while (tex){
	DocNode texture =CreateNode(textures,"texture");
	texture->SetAttribute("name",tex->name);
        if (tex->type ==0){
	  CreateNode(texture,"file",tex->file);
	}
	if (tex->type ==1){
	  CreateNode(texture,"type",tex->file);
	}
	tex=tex->next;
    }
    DocNode materials = CreateNode(node,"materials");
    tex=first;
    while (tex){
	DocNode material=CreateNode(materials,"material");
	material->SetAttribute("name",tex->name);
	CreateNode(material,"texture",tex->name);
	tex=tex->next;
    }   
}

//----------------------- main -----------------------------------
int main(int argc, char *argv[])
{
  if(argc != 3)
    synopsis(argv[0]);
  char *name_in = argv[1];
  char *name_out = argv[2];
  FILE *in, *out;
  char * worldfile ="data/grid2cs.zip";
  
  /// READ
  if((in = fopen(name_in, "rb")) == 0)
  {
    printf("Error: could not open input file '%s'.\n", name_in);
    perror(argv[0]);
    exit(1);
  }
  csRef<iObjectRegistry> object_reg= csInitializer::CreateEnvironment(argc,argv);
  if (!object_reg) return false;
  if (! csInitializer::RequestPlugins(object_reg,
				      CS_REQUEST_VFS,
				      CS_REQUEST_END)){
      return 1;
  }
  vfs= CS_QUERY_REGISTRY(object_reg, iVFS);
  if (!vfs) {
      printf("NO VFS !!!\n");
      return 2;
  }
  
  grTexInfo *texinfo = new grTexInfo(in);
  grLooksInfo *looksinfo = new grLooksInfo(in);
  grMap *map = new grMap(in);
  map->SetLooksInfo(looksinfo);
  fclose(in);

  /// CONVERT
  printf("Converting.\n");
  grSectorInfo *secinfo = new grSectorInfo();
  secinfo->Sectorize(map);
  printf("Found %d sectors.\n", secinfo->GetNumberSectors());
  // secinfo->PrintSectors();
  /// polygonizing
  grTexPlaneInfo *texplanes = new grTexPlaneInfo();
  secinfo->Polygonize(texplanes);

  /// WRITE
  #ifdef USE_OLD
  if((out = fopen(name_out, "wb")) == 0)
  {
    printf("Error: could not open output file '%s'.\n", name_out);
    perror(argv[0]);
    exit(1);
  }
  #else
  worldfile=name_out;
   
  #endif
  xml= (csPtr<iDocumentSystem>) new(csTinyDocumentSystem);
  doc= xml->CreateDocument();
  root=doc->CreateRoot();
  DocNode tempn=root->CreateNodeBefore(CS_NODE_COMMENT);
  tempn->SetValue("Created by Grid2CS");
  DocNode worldn=CreateNode(root,"world");
  WritePlugins(worldn);
  fprintf(out, "WORLD (\n");
//   texinfo->Print(out);
  // Print Plugins - new to CS  V. 0.19 
  fprintf (out, "PLUGINS (\n");
  fprintf (out,	"PLUGIN 'thing' ('crystalspace.mesh.loader.thing')\n");
  fprintf (out,	"PLUGIN 'thingFact' ('crystalspace.mesh.loader.factory.thing')\n");
  fprintf (out,	"PLUGIN 'plane' ('crystalspace.mesh.loader.thing.plane')\n");
  fprintf (out,	"PLUGIN 'bezier' ('crystalspace.mesh.loader.thing.bezier')");
  texinfo->Print(out);
fprintf (out,	")\n\n");
  // Now Planes..
  texplanes->Print(out);
  texinfo->WriteTextures(worldn);
  
  // Sector infos..
  secinfo->Print(out);
  secinfo->WriteSectors(worldn);
  // start position
  char *startsec="room";
  grVec startpos;
  GetStart(startsec, &startpos);
  fprintf(out, "  START (SECTOR('%s') POSITION( %g,%g,%g))\n", startsec, 
    startpos.x, startpos.y, startpos.z);
  fprintf(out, ")\n");
  #ifdef USE_OLD
  fclose(out);
#endif
  DocNode startn = CreateNode (worldn, "start");
  CreateNode (startn, "sector",startsec);
  DocNode attr;
  attr = CreateNode (startn, "position");
  attr->SetAttributeAsFloat ("x", startpos.x);
  attr->SetAttributeAsFloat ("y", startpos.y);
  attr->SetAttributeAsFloat ("z", startpos.z);

  vfs->Mount("/grid2cs_temp",worldfile);
  doc->Write(vfs,"/grid2cs_temp/world");
  vfs->Unmount("/grid2cs_temp",0);
  vfs->Sync();
  vfs=0;
  delete secinfo;
  delete map;
  delete looksinfo;
  delete texinfo;
  csInitializer::DestroyApplication(object_reg);
  return 0;
}
